"""
Upgraded @SheikSadi
"""
import keras.backend as K
import numpy as np

from sam_lstm.config import *
from sam_lstm.dcn_resnet import dcn_resnet
from sam_lstm.gaussian_prior import LearningPrior
from sam_lstm.attentive_convlstm import AttentiveConvLSTM

from keras.layers import Lambda, concatenate, Conv2D, UpSampling2D


def repeat(x):
    return K.reshape(
        K.repeat(K.batch_flatten(x), nb_timestep),
        (b_s, nb_timestep, 512, shape_r_gt, shape_c_gt),
    )


def repeat_shape(s):
    return (s[0], nb_timestep) + s[1:]


def upsampling(x):
    return UpSampling2D(
        size=upsampling_factor,
        interpolation="bilinear",
    )(x)


def upsampling_shape(s):
    return s[:2] + (s[2] * upsampling_factor, s[3] * upsampling_factor)


# KL-Divergence Loss
def kl_divergence(y_true, y_pred):
    max_y_pred = K.repeat_elements(
        K.expand_dims(
            K.repeat_elements(
                K.expand_dims(K.max(K.max(y_pred, axis=2), axis=2)),
                shape_r_out,
                axis=-1,
            )
        ),
        shape_c_out,
        axis=-1,
    )
    y_pred /= max_y_pred

    sum_y_true = K.repeat_elements(
        K.expand_dims(
            K.repeat_elements(
                K.expand_dims(K.sum(K.sum(y_true, axis=2), axis=2)),
                shape_r_out,
                axis=-1,
            )
        ),
        shape_c_out,
        axis=-1,
    )
    sum_y_pred = K.repeat_elements(
        K.expand_dims(
            K.repeat_elements(
                K.expand_dims(K.sum(K.sum(y_pred, axis=2), axis=2)),
                shape_r_out,
                axis=-1,
            )
        ),
        shape_c_out,
        axis=-1,
    )
    y_true /= sum_y_true + K.epsilon()
    y_pred /= sum_y_pred + K.epsilon()

    return 10 * K.sum(
        K.sum(y_true * K.log((y_true / (y_pred + K.epsilon())) + K.epsilon()), axis=-1),
        axis=-1,
    )


# Correlation Coefficient Loss
def correlation_coefficient(y_true, y_pred):
    max_y_pred = K.repeat_elements(
        K.expand_dims(
            K.repeat_elements(
                K.expand_dims(K.max(K.max(y_pred, axis=2), axis=2)),
                shape_r_out,
                axis=-1,
            )
        ),
        shape_c_out,
        axis=-1,
    )
    y_pred /= max_y_pred

    sum_y_true = K.repeat_elements(
        K.expand_dims(
            K.repeat_elements(
                K.expand_dims(K.sum(K.sum(y_true, axis=2), axis=2)),
                shape_r_out,
                axis=-1,
            )
        ),
        shape_c_out,
        axis=-1,
    )
    sum_y_pred = K.repeat_elements(
        K.expand_dims(
            K.repeat_elements(
                K.expand_dims(K.sum(K.sum(y_pred, axis=2), axis=2)),
                shape_r_out,
                axis=-1,
            )
        ),
        shape_c_out,
        axis=-1,
    )

    y_true /= sum_y_true + K.epsilon()
    y_pred /= sum_y_pred + K.epsilon()

    N = shape_r_out * shape_c_out
    sum_prod = K.sum(K.sum(y_true * y_pred, axis=2), axis=2)
    sum_x = K.sum(K.sum(y_true, axis=2), axis=2)
    sum_y = K.sum(K.sum(y_pred, axis=2), axis=2)
    sum_x_square = K.sum(K.sum(K.square(y_true), axis=2), axis=2)
    sum_y_square = K.sum(K.sum(K.square(y_pred), axis=2), axis=2)

    num = sum_prod - ((sum_x * sum_y) / N)
    den = K.sqrt(
        (sum_x_square - K.square(sum_x) / N) * (sum_y_square - K.square(sum_y) / N)
    )

    return -2 * num / den


# Normalized Scanpath Saliency Loss
def nss(y_true, y_pred):
    max_y_pred = K.repeat_elements(
        K.expand_dims(
            K.repeat_elements(
                K.expand_dims(K.max(K.max(y_pred, axis=2), axis=2)),
                shape_r_out,
                axis=-1,
            )
        ),
        shape_c_out,
        axis=-1,
    )
    y_pred /= max_y_pred
    y_pred_flatten = K.batch_flatten(y_pred)

    y_mean = K.mean(y_pred_flatten, axis=-1)
    y_mean = K.repeat_elements(
        K.expand_dims(
            K.repeat_elements(
                K.expand_dims(K.expand_dims(y_mean)), shape_r_out, axis=-1
            )
        ),
        shape_c_out,
        axis=-1,
    )

    y_std = K.std(y_pred_flatten, axis=-1)
    y_std = K.repeat_elements(
        K.expand_dims(
            K.repeat_elements(K.expand_dims(K.expand_dims(y_std)), shape_r_out, axis=-1)
        ),
        shape_c_out,
        axis=-1,
    )

    y_pred = (y_pred - y_mean) / (y_std + K.epsilon())

    return -(
        K.sum(K.sum(y_true * y_pred, axis=2), axis=2)
        / K.sum(K.sum(y_true, axis=2), axis=2)
    )


# Gaussian priors initialization
def gaussian_priors_init(shape, name=None, **kwargs):
    means = np.random.uniform(low=0.3, high=0.7, size=shape[0] // 2)
    covars = np.random.uniform(low=0.05, high=0.3, size=shape[0] // 2)
    return K.variable(np.concatenate((means, covars), axis=0), name=name)


def sam_resnet(x):
    # Dilated Convolutional Network
    dcn = dcn_resnet(input_tensor=x[0])
    conv_feat = Conv2D(512, (3, 3), padding="same", activation="relu")(dcn.output)
    # Attentive Convolutional LSTM
    att_convlstm = Lambda(repeat, repeat_shape)(conv_feat)
    att_convlstm = AttentiveConvLSTM(
        nb_filters_in=512, nb_filters_out=512, nb_filters_att=512, nb_cols=3, nb_rows=3
    )(att_convlstm)

    # Learned Prior (1)
    priors1 = LearningPrior(nb_gaussian=nb_gaussian, init=gaussian_priors_init)(x[1])
    concateneted = concatenate([att_convlstm, priors1], axis=1)
    learned_priors1 = Conv2D(
        512, (5, 5), padding="same", activation="relu", dilation_rate=(4, 4)
    )(concateneted)

    # Learned Prior (2)
    priors2 = LearningPrior(nb_gaussian=nb_gaussian, init=gaussian_priors_init)(x[1])
    concateneted = concatenate([learned_priors1, priors2], axis=1)
    learned_priors2 = Conv2D(
        512, (5, 5), padding="same", activation="relu", dilation_rate=(4, 4)
    )(concateneted)

    # Final Convolutional Layer
    outs = Conv2D(1, (1, 1), padding="same", activation="relu")(learned_priors2)
    outs_up = Lambda(upsampling, upsampling_shape)(outs)

    return [outs_up, outs_up, outs_up]



#########################################################################
# RESOURCES URL														    #
#########################################################################
DATASET_IMAGES_URL = (
    "https://github.com/SheikSadi/SAM-LSTM-RESNET/releases/download/1.0.0/images.zip"
)
DATASET_MAPS_URL = (
    "https://github.com/SheikSadi/SAM-LSTM-RESNET/releases/download/1.0.0/maps.zip"
)
DATASET_FIXS_URL = (
    "https://github.com/SheikSadi/SAM-LSTM-RESNET/releases/download/1.0.0/fixations.zip"
)
SAM_RESNET_SALICON_2017_WEIGHTS = "https://github.com/SheikSadi/SAM-LSTM-RESNET/releases/download/1.0.0/sam-resnet_salicon_weights.pkl"
TH_WEIGHTS_PATH_NO_TOP = "https://github.com/SheikSadi/SAM-LSTM-RESNET/releases/download/1.0.0/resnet50_weights_th_dim_ordering_th_kernels_notop.h5"
#########################################################################
# MODEL PARAMETERS														#
#########################################################################
# batch size
b_s = 1
# steps per epoch
steps_per_epoch = 500
# number of rows of input images
shape_r = 240
# number of cols of input images
shape_c = 320
# number of rows of downsampled maps
shape_r_gt = 30
# number of cols of downsampled maps
shape_c_gt = 40
# number of rows of model outputs
shape_r_out = 480
# number of cols of model outputs
shape_c_out = 640
# final upsampling factor
upsampling_factor = 16
# number of epochs
nb_epoch = 20
# number of timestep
nb_timestep = 4
# number of learned priors
nb_gaussian = 16
# Percentage of the total attention to retain
retained_attention = 0.4
# aspect ratio of the cropped image
aspect_ratio = 1.44  # 260/180
gaussina_sigma = 1
#########################################################################
# TRAINING SETTINGS										            	#
#########################################################################
# path of training images
dataset_path = "/mnt/d/smart-cropping/salicon_dataset_2017/"



import os
import numpy as np
from sam_lstm.config import (
    nb_gaussian,
    shape_r_gt,
    shape_c_gt,
    shape_r,
    shape_c,
    shape_r_out,
    shape_c_out,
    b_s,
)
from sam_lstm.utilities import (
    preprocess_images,
    preprocess_maps,
    preprocess_fixmaps,
)


def generator(
    images,
    maps,
    fixs,
):
    batch_size = b_s

    n_images = len(images)

    if n_images % batch_size != 0:
        raise Exception(
            f"The number (in your case: {n_images}) of training/validation images "
            "should be a multiple of the batch size. Please change the batch size."
        )

    gaussian = np.zeros((batch_size, nb_gaussian, shape_r_gt, shape_c_gt))

    counter = 0
    img_yielded = 0
    while True:
        X = (
            preprocess_images(images[counter : counter + batch_size], shape_r, shape_c),
        )
        Y = preprocess_maps(
            maps[counter : counter + batch_size], shape_r_out, shape_c_out
        )
        Y_fix = preprocess_fixmaps(
            fixs[counter : counter + batch_size], shape_r_out, shape_c_out
        )
        yield [X, gaussian], [Y, Y, Y_fix]

        img_yielded += 1
        if img_yielded == n_images:
            break
        else:
            counter = (counter + batch_size) % n_images


def generator_test(fnames, images_path):
    batch_size = b_s
    images = [os.path.join(images_path, fname) for fname in fnames]

    n_images = len(images)

    gaussian = np.zeros((batch_size, nb_gaussian, shape_r_gt, shape_c_gt))

    counter = 0
    img_yielded = 0
    while True:
        yield [
            [
                preprocess_images(
                    images[counter : counter + batch_size], shape_r, shape_c
                ),
                gaussian,
            ]
        ]
        img_yielded += 1
        if img_yielded == n_images:
            break
        else:
            counter = (counter + batch_size) % n_images



import os
import numpy as np
import cv2
import math
from skimage.feature import peak_local_max
from scipy.cluster.vq import kmeans


def find_max_subarray(array: np.ndarray, window_w: int, threshold: float) -> tuple:
    assert len(array.shape) == 1

    best_sum = -1
    start_idx = None

    array_cum = np.pad(np.cumsum(array), pad_width=[(1, 0)])

    max_start_idx = array.shape[0] - window_w

    for idx in range(max_start_idx + 1):
        cumsum_upto_windowend = array_cum[idx + window_w]
        cumsum_before_windowstart = array_cum[idx]
        subarray_sum = cumsum_upto_windowend - cumsum_before_windowstart
        if subarray_sum > threshold and subarray_sum > best_sum:
            best_sum = subarray_sum
            start_idx = idx

    return start_idx, best_sum


def find_rectangle(img_array, asp_ratio, keep_attention):
    img_h, img_w = img_array.shape
    if img_h > img_w:
        transpose = True
        array2d = img_array.T
        img_h, img_w = img_w, img_h
    else:
        transpose = False
        array2d = img_array
    array2d_hcum = np.pad(np.cumsum(array2d, axis=0), pad_width=[(1, 0), (0, 0)])
    img_h, img_w = array2d.shape
    img_area = img_h * img_w

    total_attention = np.sum(array2d)
    threshold_attention = keep_attention * total_attention

    # initialize
    y_start = 0
    min_height = 1
    y_finish = y_start + min_height
    best_area = img_area
    failedToFind = True

    while True:
        window_h = y_finish - y_start
        window_w = math.ceil(asp_ratio * window_h)

        if not (
            y_finish <= img_h
            and window_h >= min_height
            and window_h <= img_h
            and window_w <= img_w
        ):
            break

        subarray2d = array2d_hcum[y_finish] - array2d_hcum[y_start]
        x_start, attention_kept = find_max_subarray(
            subarray2d, window_w, threshold_attention
        )
        if attention_kept > 0:
            box_area = window_w * window_h
            if (box_area < best_area) or (
                box_area == best_area and attention_kept > best_attention
            ):
                best_area = box_area
                x, y, w, h = x_start, y_start, window_w, window_h
                best_attention = attention_kept
                failedToFind = False
            y_start += 1
        else:
            y_finish += 1

    if failedToFind:
        return {}
    else:
        attention_factor = best_attention / total_attention
        area_factor = w * h / img_area
        density_factor = attention_factor / area_factor
        return {
            "coords": (y, x, h, w) if transpose else (x, y, w, h),
            "area": area_factor,
            "attention": attention_factor,
            "density": density_factor,
        }


def find_best_rectangle(
    salient_ndimage, a_r, min_attention, step=0.02, alpha=10, beta=10, gamma=0.1
):
    results = {}
    attention = 1
    count = 0
    while attention >= min_attention:
        attention -= step * (2**count)
        count += 1
        result = find_rectangle(salient_ndimage, a_r, attention)
        if result:
            score = (
                -alpha * math.log10(1 - result["attention"])
                - beta * math.log10(1 - result["area"])
                + gamma ** (result["density"])
            )
            results[score] = result.pop("coords")

    if results:
        sorted_scores = sorted(results)
        print(sorted_scores)
        x, y, w, h = results[sorted_scores[-1]]
        return x, y, w, h
    else:
        raise Exception(
            f"Failed to crop with aspect ratio: {a_r}, minimum attention: {min_attention}"
        )


def get_centroids(array2d, maximum_gap=0.2, peak_theshold=0.5):
    maximum_distortion = array2d.shape[0] * maximum_gap
    for _k in [1, 2, 3, 4]:
        peaks = peak_local_max(array2d, threshold_rel=peak_theshold).astype(np.float32)
        k_peaks, distortion = kmeans(peaks.astype(float), _k)
        if distortion < maximum_distortion:
            return k_peaks.astype(np.uint32)


def descend_from_hilltop(array2d, cent_ij, alpha=1.5, beta=0.5, asp_ratio=1.44):
    cent_i, cent_j = cent_ij
    image_h, image_w = array2d.shape
    _1_pct_height = int(image_h * 0.05)
    total_area = image_h * image_w
    total_attention = array2d.sum()

    scores = []
    attentions = []
    densities = []
    coords = []

    pad_top = _1_pct_height
    pad_bottom = _1_pct_height
    while True:
        pad_right = asp_ratio * pad_bottom
        pad_left = asp_ratio * pad_top

        start_i = int(cent_i - pad_top)
        start_j = int(cent_j - pad_left)

        finish_i = int(cent_i + pad_bottom)
        finish_j = int(cent_j + pad_right)

        if start_i < 0 or finish_i >= image_h or start_j < 0 or finish_j >= image_w:
            break
        else:
            attention = array2d[start_i:finish_i, start_j:finish_j].sum()
            attention_factor = attention / total_attention
            attentions.append(attention_factor)

            area = (finish_i - start_i + 1) * (finish_j - start_j + 1)
            area_factor = area / total_area

            density_factor = attention_factor / area_factor
            densities.append(density_factor)

            coords.append([start_i, start_j, finish_i, finish_j])

            pad_bottom += _1_pct_height
            pad_top += _1_pct_height

    attentions = np.array(attentions)
    densities = np.array(densities)
    scores = np.tanh(densities**alpha) * (attentions**beta)

    start_i, start_j, finish_i, finish_j = coords[np.argmax(scores)]
    start_x, start_y, finish_x, finish_y = start_j, start_i, finish_j, finish_i

    return start_x, start_y, finish_x, finish_y


def crop(
    original_image_path,
    saliency_map_path,
    cropped_image_path,
    boxed_image_path,
    a_r,
    attention,
    **kwargs,
):
    BLUE = (255, 0, 0)
    THICKNESS = 5

    original_ndimage = cv2.imread(original_image_path, cv2.IMREAD_COLOR)
    boxed = np.copy(original_ndimage)
    salient_ndimage = cv2.imread(saliency_map_path, cv2.IMREAD_GRAYSCALE)

    # x, y, w, h = find_best_rectangle(salient_ndimage, a_r, attention, **kwargs)
    # cropped_ndimage = original_ndimage[y : y + h, x : x + w, :]
    # crop_success = cv2.imwrite(cropped_image_path, cropped_ndimage.astype(int))
    # Draw a diagonal blue line with thickness of 5 px
    # cv2.rectangle(original_ndimage, (x, y), (x + w, y + h), blue, thickness)
    for i, cent_ij in enumerate(get_centroids(salient_ndimage)):
        if i > 0:
            name, ext = cropped_image_path.rsplit(".", 1)
            cropped_image_path = f"{name}_{i}.{ext}"

        start_x, start_y, finish_x, finish_y = descend_from_hilltop(
            salient_ndimage, cent_ij
        )
        cropped_ndimage = original_ndimage[start_y:finish_y, start_x:finish_x, :]
        cv2.imwrite(cropped_image_path, cropped_ndimage)
        cv2.rectangle(boxed, (start_x, start_y), (finish_x, finish_y), BLUE, THICKNESS)
    cv2.imwrite(boxed_image_path, boxed)


def batch_crop_images(
    originals_folder,
    maps_folder,
    crops_folder,
    boxes_folder,
    aspect_ratio,
    retained_attention,
    **kwargs,
):
    for _dir in [crops_folder, boxes_folder]:
        if not os.path.exists(_dir):
            os.mkdir(_dir)

    for fname in os.listdir(maps_folder):
        if fname in os.listdir(originals_folder):
            original_file = os.path.join(originals_folder, fname)
            mapping_file = os.path.join(maps_folder, fname)
            crop_file = os.path.join(crops_folder, fname)
            box_file = os.path.join(boxes_folder, fname)
            try:
                crop_success = crop(
                    original_file,
                    mapping_file,
                    crop_file,
                    box_file,
                    aspect_ratio,
                    retained_attention,
                    **kwargs,
                )
            except Exception as e:
                print(f"{e} for {fname}")
                continue
            else:
                print(f"Cropped and boxed {fname} successfully")



"""
Upgraded @SheikSadi
"""
import tensorflow as tf
import numpy as np
import keras.backend as K
from keras.layers import Layer, InputSpec
from keras import initializers, regularizers, constraints


floatX = K.floatx()


class LearningPrior(Layer):
    def __init__(
        self,
        nb_gaussian,
        init="normal",
        weights=None,
        W_regularizer=None,
        activity_regularizer=None,
        W_constraint=None,
        **kwargs,
    ):
        self.nb_gaussian = nb_gaussian
        self.init = initializers.get(init)

        self.W_regularizer = regularizers.get(W_regularizer)
        self.activity_regularizer = regularizers.get(activity_regularizer)

        self.W_constraint = constraints.get(W_constraint)

        self.input_spec = [InputSpec(ndim=4)]
        self.initial_weights = weights
        super(LearningPrior, self).__init__(**kwargs)

    def build(self, input_shape):
        if self.initial_weights is None:
            self.W = self.add_weight(
                name=f"{self.name}_W",
                shape=(self.nb_gaussian * 4,),
                initializer=self.init,
                regularizer=self.W_regularizer,
                trainable=True,
                constraint=self.W_constraint,
            )
        else:
            self.set_weights(self.initial_weights)
            del self.initial_weights

    def get_output_shape_for(self, input_shape):
        self.b_s = input_shape[0]
        self.height = input_shape[2]
        self.width = input_shape[3]

        return self.b_s, self.nb_gaussian, self.height, self.width

    def call(self, x, mask=None):
        mu_x = K.clip(self.W[: self.nb_gaussian], 0.25, 0.75)
        mu_y = K.clip(self.W[self.nb_gaussian : self.nb_gaussian * 2], 0.35, 0.65)
        sigma_x = K.clip(self.W[self.nb_gaussian * 2 : self.nb_gaussian * 3], 0.1, 0.9)
        sigma_y = K.clip(self.W[self.nb_gaussian * 3 :], 0.2, 0.8)

        self.b_s = x.shape[0] if x.shape[0] else 1
        self.height = x.shape[2]
        self.width = x.shape[3]

        e = self.height / self.width
        e1 = (1 - e) / 2
        e2 = (1 + e) / 2

        x_t, y_t = tf.meshgrid(
            tf.cast(tf.linspace(0, 1, self.width), dtype=tf.float32),
            tf.cast(tf.linspace(e1, e2, self.height), dtype=tf.float32),
        )

        x_t = tf.repeat(tf.expand_dims(x_t, axis=2), self.nb_gaussian, axis=2)
        y_t = tf.repeat(tf.expand_dims(y_t, axis=2), self.nb_gaussian, axis=2)

        gaussian = (
            1
            / (2 * np.pi * sigma_x * sigma_y + K.epsilon())
            * tf.math.exp(
                -(
                    (x_t - mu_x) ** 2 / (2 * sigma_x**2 + K.epsilon())
                    + (y_t - mu_y) ** 2 / (2 * sigma_y**2 + K.epsilon())
                )
            )
        )

        gaussian /= tf.math.reduce_sum(gaussian, axis=[0, 1])

        gaussian = tf.repeat(tf.expand_dims(gaussian, axis=0), self.b_s, axis=0)

        output = tf.transpose(gaussian, perm=[0, 3, 1, 2])  # To NCHW format

        return output

    def get_config(self):
        config = {
            "nb_gaussian": self.nb_gaussian,
            "init": self.init.__name__,
            "W_regularizer": self.W_regularizer.get_config()
            if self.W_regularizer
            else None,
            "activity_regularizer": self.activity_regularizer.get_config()
            if self.activity_regularizer
            else None,
            "W_constraint": self.W_constraint.get_config()
            if self.W_constraint
            else None,
        }
        base_config = super(LearningPrior, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))



"""
This code is part of the Keras ResNet-50 model
"""
import keras.backend as K
from keras.layers import (
    add,
    Input,
    Activation,
    Conv2D,
    MaxPooling2D,
    ZeroPadding2D,
    BatchNormalization,
)
from keras.models import Model
from keras.utils import get_file
from sam_lstm.config import TH_WEIGHTS_PATH_NO_TOP


def identity_block(input_tensor, kernel_size, filters, stage, block):
    """
    UPGRADED - @SheikSadi
    """
    nb_filter1, nb_filter2, nb_filter3 = filters
    bn_axis = 1

    conv_name_base = "res" + str(stage) + block + "_branch"
    bn_name_base = "bn" + str(stage) + block + "_branch"

    x = Conv2D(
        nb_filter1, kernel_size=(1, 1), name=conv_name_base + "2a", trainable=False
    )(input_tensor)
    x = BatchNormalization(axis=bn_axis, name=bn_name_base + "2a", trainable=False)(x)
    x = Activation("relu", trainable=False)(x)

    x = Conv2D(
        nb_filter2,
        kernel_size,
        padding="same",
        name=conv_name_base + "2b",
        trainable=False,
    )(x)
    x = BatchNormalization(axis=bn_axis, name=bn_name_base + "2b", trainable=False)(x)
    x = Activation("relu", trainable=False)(x)

    x = Conv2D(
        nb_filter3, kernel_size=(1, 1), name=conv_name_base + "2c", trainable=False
    )(x)
    x = BatchNormalization(axis=bn_axis, name=bn_name_base + "2c", trainable=False)(x)

    x = add([x, input_tensor])
    x = Activation("relu", trainable=False)(x)
    return x


def conv_block(input_tensor, kernel_size, filters, stage, block, strides=(2, 2)):
    """
    UPGRADED - @SheikSadi
    """
    nb_filter1, nb_filter2, nb_filter3 = filters
    bn_axis = 1

    conv_name_base = "res" + str(stage) + block + "_branch"
    bn_name_base = "bn" + str(stage) + block + "_branch"

    x = Conv2D(
        nb_filter1,
        kernel_size=(1, 1),
        strides=strides,
        name=conv_name_base + "2a",
        trainable=False,
    )(input_tensor)
    x = BatchNormalization(axis=bn_axis, name=bn_name_base + "2a", trainable=False)(x)
    x = Activation("relu", trainable=False)(x)

    x = Conv2D(
        nb_filter2,
        kernel_size,
        padding="same",
        name=conv_name_base + "2b",
        trainable=False,
    )(x)
    x = BatchNormalization(axis=bn_axis, name=bn_name_base + "2b", trainable=False)(x)
    x = Activation("relu", trainable=False)(x)

    x = Conv2D(
        nb_filter3, kernel_size=(1, 1), name=conv_name_base + "2c", trainable=False
    )(x)
    x = BatchNormalization(axis=bn_axis, name=bn_name_base + "2c", trainable=False)(x)

    shortcut = Conv2D(
        nb_filter3,
        kernel_size=(1, 1),
        strides=strides,
        name=conv_name_base + "1",
        trainable=False,
    )(input_tensor)
    shortcut = BatchNormalization(
        axis=bn_axis, name=bn_name_base + "1", trainable=False
    )(shortcut)

    x = add([x, shortcut])
    x = Activation("relu", trainable=False)(x)
    return x


def conv_block_atrous(
    input_tensor, kernel_size, filters, stage, block, dilation_rate=(2, 2)
):
    """
    UPGRADED: @SheikSadi
    """
    nb_filter1, nb_filter2, nb_filter3 = filters
    bn_axis = 1

    conv_name_base = "res" + str(stage) + block + "_branch"
    bn_name_base = "bn" + str(stage) + block + "_branch"

    x = Conv2D(nb_filter1, (1, 1), name=conv_name_base + "2a", trainable=False)(
        input_tensor
    )
    x = BatchNormalization(axis=bn_axis, name=bn_name_base + "2a", trainable=False)(x)
    x = Activation("relu", trainable=False)(x)

    x = Conv2D(
        nb_filter2,
        kernel_size,
        padding="same",
        dilation_rate=dilation_rate,
        name=conv_name_base + "2b",
        trainable=False,
    )(x)
    x = BatchNormalization(axis=bn_axis, name=bn_name_base + "2b", trainable=False)(x)
    x = Activation("relu", trainable=False)(x)

    x = Conv2D(nb_filter3, (1, 1), name=conv_name_base + "2c", trainable=False)(x)
    x = BatchNormalization(axis=bn_axis, name=bn_name_base + "2c", trainable=False)(x)

    shortcut = Conv2D(nb_filter3, (1, 1), name=conv_name_base + "1", trainable=False)(
        input_tensor
    )
    shortcut = BatchNormalization(
        axis=bn_axis, name=bn_name_base + "1", trainable=False
    )(shortcut)

    x = add([x, shortcut])
    x = Activation("relu", trainable=False)(x)
    return x


def identity_block_atrous(
    input_tensor, kernel_size, filters, stage, block, dilation_rate=(2, 2)
):
    """
    UPGRADED - @SheikSadi
    """
    nb_filter1, nb_filter2, nb_filter3 = filters
    bn_axis = 1

    conv_name_base = "res" + str(stage) + block + "_branch"
    bn_name_base = "bn" + str(stage) + block + "_branch"

    x = Conv2D(
        nb_filter1, kernel_size=(1, 1), name=conv_name_base + "2a", trainable=False
    )(input_tensor)
    x = BatchNormalization(axis=bn_axis, name=bn_name_base + "2a", trainable=False)(x)
    x = Activation("relu", trainable=False)(x)

    x = Conv2D(
        nb_filter2,
        kernel_size,
        dilation_rate=dilation_rate,
        padding="same",
        name=conv_name_base + "2b",
        trainable=False,
    )(x)
    x = BatchNormalization(axis=bn_axis, name=bn_name_base + "2b", trainable=False)(x)
    x = Activation("relu", trainable=False)(x)

    x = Conv2D(nb_filter3, (1, 1), name=conv_name_base + "2c", trainable=False)(x)
    x = BatchNormalization(axis=bn_axis, name=bn_name_base + "2c", trainable=False)(x)

    x = add([x, input_tensor])
    x = Activation("relu", trainable=False)(x)
    return x


def dcn_resnet(input_tensor=None):
    input_shape = (3, None, None)

    if input_tensor is None:
        img_input = Input(shape=input_shape)
    else:
        if not K.is_keras_tensor(input_tensor):
            img_input = Input(tensor=input_tensor)
        else:
            img_input = input_tensor

    bn_axis = 1

    # conv_1
    x = ZeroPadding2D((3, 3), trainable=False)(img_input)
    x = Conv2D(64, (7, 7), strides=(2, 2), name="conv1", trainable=False)(x)
    x = BatchNormalization(axis=bn_axis, name="bn_conv1", trainable=False)(x)
    x = Activation("relu", trainable=False)(x)
    x = MaxPooling2D((3, 3), strides=(2, 2), padding="same", trainable=False)(x)

    # conv_2
    x = conv_block(x, 3, [64, 64, 256], stage=2, block="a", strides=(1, 1))
    x = identity_block(x, 3, [64, 64, 256], stage=2, block="b")
    x = identity_block(x, 3, [64, 64, 256], stage=2, block="c")

    # conv_3
    x = conv_block(x, 3, [128, 128, 512], stage=3, block="a", strides=(2, 2))
    x = identity_block(x, 3, [128, 128, 512], stage=3, block="b")
    x = identity_block(x, 3, [128, 128, 512], stage=3, block="c")
    x = identity_block(x, 3, [128, 128, 512], stage=3, block="d")

    # conv_4
    x = conv_block_atrous(
        x, 3, [256, 256, 1024], stage=4, block="a", dilation_rate=(2, 2)
    )
    x = identity_block_atrous(
        x, 3, [256, 256, 1024], stage=4, block="b", dilation_rate=(2, 2)
    )
    x = identity_block_atrous(
        x, 3, [256, 256, 1024], stage=4, block="c", dilation_rate=(2, 2)
    )
    x = identity_block_atrous(
        x, 3, [256, 256, 1024], stage=4, block="d", dilation_rate=(2, 2)
    )
    x = identity_block_atrous(
        x, 3, [256, 256, 1024], stage=4, block="e", dilation_rate=(2, 2)
    )
    x = identity_block_atrous(
        x, 3, [256, 256, 1024], stage=4, block="f", dilation_rate=(2, 2)
    )

    # conv_5
    x = conv_block_atrous(
        x, 3, [512, 512, 2048], stage=5, block="a", dilation_rate=(4, 4)
    )
    x = identity_block_atrous(
        x, 3, [512, 512, 2048], stage=5, block="b", dilation_rate=(4, 4)
    )
    x = identity_block_atrous(
        x, 3, [512, 512, 2048], stage=5, block="c", dilation_rate=(4, 4)
    )

    # Create model
    model = Model(img_input, x)

    # Load weights
    weights_path = get_file(
        "resnet50_weights_th_dim_ordering_th_kernels_notop.h5",
        TH_WEIGHTS_PATH_NO_TOP,
        cache_subdir="weights",
        file_hash="f64f049c92468c9affcd44b0976cdafe",
    )
    model.load_weights(weights_path)

    return model



import tensorflow as tf
import keras.backend as K
from keras.layers import Layer, InputSpec
from keras import initializers


class AttentiveConvLSTM(Layer):
    """
    att_convlstm = AttentiveConvLSTM(
        nb_filters_in=512, nb_filters_out=512, nb_filters_att=512, nb_cols=3, nb_rows=3
    )(att_convlstm)
    """

    def __init__(
        self,
        nb_filters_in,
        nb_filters_out,
        nb_filters_att,
        nb_rows,
        nb_cols,
        init="normal",
        inner_init="orthogonal",
        attentive_init="zero",
        **kwargs,
    ):
        self.nb_filters_in = nb_filters_in
        self.nb_filters_out = nb_filters_out
        self.nb_filters_att = nb_filters_att
        self.nb_rows = nb_rows
        self.nb_cols = nb_cols
        self.init = init
        self.inner_init = inner_init
        self.attentive_init = attentive_init
        super().__init__(**kwargs)

    def get_output_shape_for(self, input_shape):
        return input_shape[:1] + (self.nb_filters_out,) + input_shape[3:]

    def compute_mask(self, input, mask):
        return None

    def get_initial_states(self, x):
        # we will skip the time-axis (axis=1)
        initial_state_shape = (x.shape[0], x.shape[2], x.shape[3], x.shape[4])
        initial_state = tf.zeros(initial_state_shape)
        initial_states = [initial_state for _ in self.states]

        return initial_states

    def build(self, input_shape):
        self.inputspec = [InputSpec(shape=input_shape)]
        self.states = [None, None]
        in_channels = input_shape[2]

        def init_conv_weights(out_channels, name, init, bias=True):
            _kernel = self.add_weight(
                f"{name}_kernel",
                shape=[out_channels, in_channels, self.nb_rows, self.nb_cols],
                trainable=True,
                initializer=initializers.get(init),
            )
            if bias:
                _bias = self.add_weight(
                    f"{name}_bias",
                    shape=[out_channels],
                    trainable=True,
                    initializer=initializers.get(self.init),
                )
            else:
                _bias = tf.zeros(shape=[out_channels])
            return _kernel, _bias

        self.Wa = init_conv_weights(self.nb_filters_att, "Wa", self.init)
        self.Ua = init_conv_weights(self.nb_filters_att, "Ua", self.init)
        self.Va = init_conv_weights(1, "Va", self.attentive_init, bias=False)

        self.Wi = init_conv_weights(self.nb_filters_out, "Wi", self.init)
        self.Ui = init_conv_weights(self.nb_filters_out, "Ui", self.inner_init)

        self.Wf = init_conv_weights(self.nb_filters_out, "Wf", self.init)
        self.Uf = init_conv_weights(self.nb_filters_out, "Uf", self.inner_init)

        self.Wc = init_conv_weights(self.nb_filters_out, "Wc", self.init)
        self.Uc = init_conv_weights(self.nb_filters_out, "Uc", self.inner_init)

        self.Wo = init_conv_weights(self.nb_filters_out, "Wo", self.init)
        self.Uo = init_conv_weights(self.nb_filters_out, "Uo", self.inner_init)

    def preprocess_input(self, x):
        return x

    @staticmethod
    def conv(kernel_bias, input):
        kernel, bias_ = kernel_bias
        kernelT = tf.transpose(kernel, perm=[3, 2, 1, 0])
        kernelT_x_input = tf.nn.conv2d(
            input, kernelT, strides=1, padding="SAME", data_format="NCHW"
        )

        bias_ = tf.expand_dims(tf.expand_dims(bias_, axis=1), axis=2)
        bias = tf.repeat(
            tf.repeat(bias_, repeats=input.shape[2], axis=1),
            repeats=input.shape[3],
            axis=2,
        )
        return kernelT_x_input + bias

    def step(self, X, states):
        sigmoid = tf.math.sigmoid
        tanh = tf.math.tanh
        conv = self.conv

        Ht_1 = states[0]
        Ct_1 = states[1]

        Zt = conv(self.Va, tanh(conv(self.Wa, X) + conv(self.Ua, Ht_1)))
        At = tf.repeat(
            tf.reshape(
                tf.nn.softmax(K.batch_flatten(Zt)),
                (X.shape[0], 1, X.shape[2], X.shape[3]),
            ),
            repeats=X.shape[1],
            axis=1,
        )

        Xt_ = X * At

        It = sigmoid(conv(self.Wi, Xt_) + conv(self.Ui, Ht_1))
        Ft = sigmoid(conv(self.Wf, Xt_) + conv(self.Uf, Ht_1))
        Ot = sigmoid(conv(self.Wo, Xt_) + conv(self.Uo, Ht_1))
        Gt = tanh(conv(self.Wc, Xt_) + conv(self.Uc, Ht_1))
        Ct = Ft * Ct_1 + It * Gt
        Ht = Ot * tanh(Ct)

        return Ht, [Ht, Ct]

    def get_constants(self, x):
        return []

    def call(self, x, mask=None):
        initial_states = self.get_initial_states(x)
        constants = self.get_constants(x)
        preprocessed_input = self.preprocess_input(x)

        last_output, outputs, states = K.rnn(
            self.step,
            preprocessed_input,
            initial_states,
            go_backwards=False,
            mask=mask,
            constants=constants,
            unroll=False,
            input_length=x.shape[1],
        )

        if len(last_output.shape) == 3:
            last_output = tf.expand_dims(last_output, axis=0)

        return last_output

    def get_config(self):
        config_to_serialize = dict(
            nb_filters_in=self.nb_filters_in,
            nb_filters_out=self.nb_filters_out,
            nb_filters_att=self.nb_filters_att,
            nb_rows=self.nb_rows,
            nb_cols=self.nb_cols,
            init=self.init,
            inner_init=self.inner_init,
            attentive_init=self.attentive_init,
        )
        config = super().get_config()
        config.update(config_to_serialize)
        return config



import cv2
import os

import tensorflow as tf
from keras.optimizers import RMSprop

from keras.callbacks import EarlyStopping, ModelCheckpoint
from keras.layers import Input
from keras.models import Model
from keras.utils import get_file

from sam_lstm.utilities import postprocess_predictions
from sam_lstm.models import (
    kl_divergence,
    correlation_coefficient,
    nss,
    sam_resnet,
)
from sam_lstm.generator import generator, generator_test
from sam_lstm.cropping import batch_crop_images
from sam_lstm.config import (
    shape_r,
    shape_c,
    shape_r_gt,
    shape_c_gt,
    nb_gaussian,
    b_s,
    nb_epoch,
    steps_per_epoch,
    aspect_ratio,
    retained_attention,
    DATASET_IMAGES_URL,
    DATASET_MAPS_URL,
    DATASET_FIXS_URL,
    SAM_RESNET_SALICON_2017_WEIGHTS,
    dataset_path,
    gaussina_sigma,
)


tf.keras.backend.set_image_data_format(data_format="channels_first")


class SalMap:
    @classmethod
    def auto(cls):
        salmap = cls()
        salmap.compile()
        salmap.load_weights()
        salmap.predict_maps()
        salmap.box_and_crop()

    def __init__(self):
        self.x = Input((3, shape_r, shape_c))
        self.x_maps = Input((nb_gaussian, shape_r_gt, shape_c_gt))

    def compile(self):
        self.model = Model(
            inputs=[self.x, self.x_maps], outputs=sam_resnet([self.x, self.x_maps])
        )
        self.model.compile(
            RMSprop(learning_rate=1e-4),
            loss=[kl_divergence, correlation_coefficient, nss],
            loss_weights=[10, -2, -1],
        )

    @staticmethod
    def get_valid_images(imgs_path, maps_path, fixs_path):
        _images = {
            fname.rsplit(".", 1)[0]: os.path.join(imgs_path, fname)
            for fname in os.listdir(imgs_path)
            if fname.endswith((".jpg", ".jpeg", ".png"))
        }
        _maps = {
            fname.rsplit(".", 1)[0]: os.path.join(maps_path, fname)
            for fname in os.listdir(maps_path)
            if fname.endswith((".jpg", ".jpeg", ".png"))
        }

        _fixs = {
            fname.rsplit(".", 1)[0]: os.path.join(fixs_path, fname)
            for fname in os.listdir(fixs_path)
            if fname.endswith(".mat")
        }

        images = []
        maps = []
        fixs = []

        # make sure all files in images have corresponding files in maps and fixs
        for fname in set(_images).intersection(_maps, _fixs):
            images.append(_images[fname])
            maps.append(_maps[fname])
            fixs.append(_fixs[fname])

        del _images, _maps, _fixs

        return images, maps, fixs

    def train(
        self,
        dataset_path=dataset_path,
        checkpoint_path="weights/",
        steps_per_epoch=steps_per_epoch,
    ):
        imgs_path = os.path.join(dataset_path, "images")
        maps_path = os.path.join(dataset_path, "maps")
        fixs_path = os.path.join(dataset_path, "fixations")

        if not os.path.exists(imgs_path):
            print(f"Didn't find {imgs_path}")
            imgs_zip_path = get_file(
                origin=DATASET_IMAGES_URL,
                file_hash="2c72253ccb5288118864ebd2ab15a55e",
                hash_algorithm="md5",
                extract=True,
                cache_dir=os.getcwd(),
                cache_subdir=dataset_path,
            )
            imgs_path = imgs_zip_path.replace(".zip", "")

        if not os.path.exists(maps_path):
            print(f"Didn't find {maps_path}")
            maps_zip_path = get_file(
                origin=DATASET_MAPS_URL,
                file_hash="5218595acfeec3b9fc0a4964d0566360",
                hash_algorithm="md5",
                extract=True,
                cache_dir=os.getcwd(),
                cache_subdir=dataset_path,
            )
            maps_path = maps_zip_path.replace(".zip", "")

        if not os.path.exists(fixs_path):
            print(f"Didn't find {fixs_path}")
            fixs_zip_path = get_file(
                origin=DATASET_FIXS_URL,
                file_hash="0d6f4a54c3d36ccc85a74b1b4b40bed5",
                hash_algorithm="md5",
                extract=True,
                cache_dir=os.getcwd(),
                cache_subdir=dataset_path,
            )
            fixs_path = fixs_zip_path.replace(".zip", "")

        imgs_train_path = os.path.join(imgs_path, "train")
        maps_train_path = os.path.join(maps_path, "train")
        fixs_train_path = os.path.join(fixs_path, "train")

        imgs_val_path = os.path.join(imgs_path, "val")
        maps_val_path = os.path.join(maps_path, "val")
        fixs_val_path = os.path.join(fixs_path, "val")

        for path in [
            imgs_train_path,
            maps_train_path,
            fixs_train_path,
            imgs_val_path,
            maps_val_path,
            fixs_val_path,
        ]:
            if not os.path.exists(path):
                raise Exception(f"Didn't find the {path}! Can't start training...")

        imgs_train, maps_train, fixs_train = self.get_valid_images(
            imgs_train_path, maps_train_path, fixs_train_path
        )

        print("Training SAM-ResNet")
        train_gen = generator(
            imgs_train,
            maps_train,
            fixs_train,
        )

        imgs_val, maps_val, fixs_val = self.get_valid_images(
            imgs_val_path, maps_val_path, fixs_val_path
        )
        validation_gen = generator(
            imgs_val,
            maps_val,
            fixs_val,
        )
        if not os.path.exists(checkpoint_path):
            raise Exception(
                f"Directory: {checkpoint_path} not found, first make sure it exists. Then, try again!"
            )

        self.model.fit(
            train_gen,
            batch_size=b_s,
            epochs=nb_epoch,
            # all training will be visited twice
            steps_per_epoch=steps_per_epoch,
            validation_data=validation_gen,
            validation_steps=int(steps_per_epoch / 3),
            callbacks=[
                EarlyStopping(patience=3),
                ModelCheckpoint(
                    os.path.join(checkpoint_path, "sam-resnet-salicon.h5"),
                    monitor="val_loss",
                    verbose=0,
                    save_best_only=True,
                    save_weights_only=True,
                    mode="auto",
                    save_freq="epoch",
                ),
            ],
        )

    def load_weights(self, weights_dir=None):
        if weights_dir:
            if not os.path.exists(weights_dir):
                weights_dir = None
        else:
            fname = "sam-resnet_salicon_weights.pkl"
            cache_subdir = "weights"

            weights_dir = get_file(
                fname,
                SAM_RESNET_SALICON_2017_WEIGHTS,
                cache_subdir=cache_subdir,
                file_hash="92b5f89fd34a3968776a5c4327efb32c",
            )

        self.model.load_weights(weights_dir)

    def predict_maps(self, imgs_test_path="/samples", sigma=gaussina_sigma):
        if imgs_test_path.startswith("/"):
            imgs_test_path = imgs_test_path.rsplit("/", 1)[1]
        # Output Folder Path
        if os.path.exists(imgs_test_path):
            self.imgs_test_path = imgs_test_path
        else:
            self.imgs_test_path = os.path.join(os.getcwd(), imgs_test_path)
        if not os.path.exists(self.imgs_test_path):
            raise Exception(
                f"Couldn't find the directory {imgs_test_path} or {self.imgs_test_path}"
            )

        home_dir = os.path.dirname(self.imgs_test_path)
        maps_folder = os.path.join(home_dir, "maps")
        if not os.path.exists(maps_folder):
            os.mkdir(maps_folder)
        cmaps_folder = os.path.join(home_dir, "cmaps")
        if not os.path.exists(cmaps_folder):
            os.mkdir(cmaps_folder)

        file_names = [
            fname
            for fname in os.listdir(self.imgs_test_path)
            if fname.endswith((".jpg", ".jpeg", ".png"))
        ]
        file_names.sort()

        print("Predicting saliency maps for " + self.imgs_test_path)
        predictions = self.model.predict(
            generator_test(file_names, imgs_test_path), steps=len(file_names)
        )[0]

        for pred, fname in zip(predictions, file_names):
            image_path = os.path.join(self.imgs_test_path, fname)
            original_image = cv2.imread(image_path, cv2.IMREAD_COLOR)
            mapped_image = postprocess_predictions(
                pred[0], original_image.shape[0], original_image.shape[1]
            )
            map_path = os.path.join(maps_folder, fname)
            cv2.imwrite(map_path, mapped_image)

            gray_map = cv2.imread(map_path, cv2.IMREAD_GRAYSCALE)
            colored_map = cv2.applyColorMap(gray_map, cv2.COLORMAP_JET)
            overlaid = cv2.addWeighted(original_image, 0.5, colored_map, 0.5, 0)
            cmap_path = os.path.join(cmaps_folder, fname)
            cv2.imwrite(cmap_path, overlaid)

    @staticmethod
    def box_and_crop(
        asp_ratio=aspect_ratio,
        threshold_attention=retained_attention,
        originals_folder="samples",
        alpha=20,
        beta=15,
        gamma=5,
    ):
        """
        score = alpha ** (% of total attention) + beta ** (% of total area) + gamma ** (attention density)
        """
        maps_folder = os.path.join(os.path.dirname(originals_folder), "maps")
        if not os.path.exists(maps_folder):
            raise Exception(
                f"Saliency mappings for the images in {originals_folder}"
                " must be present in {maps_folder}.\n"
                "Run this command - salmap.test(imgs_test_path=<original-images>) and try again!"
            )
        crops_folder = os.path.join(os.path.dirname(originals_folder), "crops")
        boxes_folder = os.path.join(os.path.dirname(originals_folder), "boxes")

        batch_crop_images(
            originals_folder,
            maps_folder,
            crops_folder,
            boxes_folder,
            asp_ratio,
            threshold_attention,
            alpha=alpha,
            beta=beta,
            gamma=gamma,
        )



import cv2
import numpy as np
import scipy.io
import scipy.ndimage
from sam_lstm.config import gaussina_sigma


def padding(img, shape_r=240, shape_c=320, channels=3):
    img_padded = np.zeros((shape_r, shape_c, channels), dtype=np.uint8)
    if channels == 1:
        img_padded = np.zeros((shape_r, shape_c), dtype=np.uint8)

    original_shape = img.shape
    rows_rate = original_shape[0] / shape_r
    cols_rate = original_shape[1] / shape_c

    if rows_rate > cols_rate:
        new_cols = (original_shape[1] * shape_r) // original_shape[0]
        img = cv2.resize(img, (new_cols, shape_r))
        if new_cols > shape_c:
            new_cols = shape_c
        img_padded[
            :,
            ((img_padded.shape[1] - new_cols) // 2) : (
                (img_padded.shape[1] - new_cols) // 2 + new_cols
            ),
        ] = img
    else:
        new_rows = (original_shape[0] * shape_c) // original_shape[1]
        img = cv2.resize(img, (shape_c, new_rows))
        if new_rows > shape_r:
            new_rows = shape_r
        img_padded[
            ((img_padded.shape[0] - new_rows) // 2) : (
                (img_padded.shape[0] - new_rows) // 2 + new_rows
            ),
            :,
        ] = img

    return img_padded


def resize_fixation(img, rows=480, cols=640):
    out = np.zeros((rows, cols))
    factor_scale_r = rows / img.shape[0]
    factor_scale_c = cols / img.shape[1]

    coords = np.argwhere(img)
    for coord in coords:
        r = int(np.round(coord[0] * factor_scale_r))
        c = int(np.round(coord[1] * factor_scale_c))
        if r == rows:
            r -= 1
        if c == cols:
            c -= 1
        out[r, c] = 1

    return out


def padding_fixation(img, shape_r=480, shape_c=640):
    img_padded = np.zeros((shape_r, shape_c))
    original_shape = img.shape

    if original_shape == img_padded.shape:
        return img
    else:
        rows_rate = original_shape[0] / shape_r
        cols_rate = original_shape[1] / shape_c

        if rows_rate > cols_rate:
            new_cols = (original_shape[1] * shape_r) // original_shape[0]
            img = resize_fixation(img, rows=shape_r, cols=new_cols)
            if new_cols > shape_c:
                new_cols = shape_c
            img_padded[
                :,
                ((img_padded.shape[1] - new_cols) // 2) : (
                    (img_padded.shape[1] - new_cols) // 2 + new_cols
                ),
            ] = img
        else:
            new_rows = (original_shape[0] * shape_c) // original_shape[1]
            img = resize_fixation(img, rows=new_rows, cols=shape_c)
            if new_rows > shape_r:
                new_rows = shape_r
            img_padded[
                ((img_padded.shape[0] - new_rows) // 2) : (
                    (img_padded.shape[0] - new_rows) // 2 + new_rows
                ),
                :,
            ] = img

        return img_padded


def preprocess_images(paths, shape_r, shape_c):
    ims = np.zeros((len(paths), shape_r, shape_c, 3))

    for i, path in enumerate(paths):
        original_image = cv2.imread(path, cv2.IMREAD_COLOR)
        padded_image = padding(original_image, shape_r, shape_c, 3)
        ims[i] = padded_image

    ims[:, :, :, 0] -= 103.939
    ims[:, :, :, 1] -= 116.779
    ims[:, :, :, 2] -= 123.68
    ims = ims.transpose((0, 3, 1, 2))

    return ims


def preprocess_maps(paths, shape_r, shape_c):
    ims = np.zeros((len(paths), 1, shape_r, shape_c))

    for i, path in enumerate(paths):
        original_map = cv2.imread(path, cv2.IMREAD_GRAYSCALE)
        padded_map = padding(original_map, shape_r, shape_c, 1)
        ims[i, 0] = padded_map.astype(np.float32)
        ims[i, 0] /= 255.0

    return ims


def preprocess_fixmaps(paths, shape_r, shape_c):
    ims = np.zeros((len(paths), 1, shape_r, shape_c))

    for i, path in enumerate(paths):
        matdata = scipy.io.loadmat(path)
        imshape = matdata["resolution"][0]
        fix_map = np.zeros(imshape)

        fixations = np.vstack([fixation[0][2] for fixation in matdata["gaze"]])
        xs = fixations[:, 0] - 1
        ys = fixations[:, 1] - 1
        fix_map[ys, xs] = 1

        ims[i, 0] = padding_fixation(fix_map, shape_r=shape_r, shape_c=shape_c)

    return ims


def postprocess_predictions(pred, shape_r, shape_c, sigma=gaussina_sigma):
    predictions_shape = pred.shape
    rows_rate = shape_r / predictions_shape[0]
    cols_rate = shape_c / predictions_shape[1]

    pred = pred / np.max(pred) * 255

    if rows_rate > cols_rate:
        new_cols = (predictions_shape[1] * shape_r) // predictions_shape[0]
        pred = cv2.resize(pred, (new_cols, shape_r))
        img = pred[
            :,
            ((pred.shape[1] - shape_c) // 2) : (
                (pred.shape[1] - shape_c) // 2 + shape_c
            ),
        ]
    else:
        new_rows = (predictions_shape[0] * shape_c) // predictions_shape[1]
        pred = cv2.resize(pred, (shape_c, new_rows))
        img = pred[
            ((pred.shape[0] - shape_r) // 2) : (
                (pred.shape[0] - shape_r) // 2 + shape_r
            ),
            :,
        ]

    img = scipy.ndimage.gaussian_filter(img, sigma=sigma)
    img = img / np.max(img) * 255

    return img.clip(0, 255).astype(np.int32)



from setuptools import setup

setup(
    name="sam_lstm",
    url="https://github.com/SheikSadi/SAM-LSTM-RESNET.git",
    author="SheikSadi",
    author_email="biis.saadi@gmail.com",
    include_package_data=True,
    packages=["sam_lstm"],
    version="1.0.1",
    description="Python library to generate saliency maps and smart cropping images",
)
